from argparse import ArgumentParser
from collections.abc import Iterable
from pathlib import Path

from unhacs.git import get_repo_tags
from unhacs.packages import DEFAULT_HASS_CONFIG_PATH
from unhacs.packages import DEFAULT_PACKAGE_FILE
from unhacs.packages import Package
from unhacs.packages import PackageType
from unhacs.packages import get_installed_packages
from unhacs.packages import read_lock_packages
from unhacs.packages import write_lock_packages


def create_parser():
    parser = ArgumentParser(
        description="Unhacs - Command line interface for the Home Assistant Community Store"
    )
    parser.add_argument(
        "--config",
        "-c",
        type=Path,
        default=DEFAULT_HASS_CONFIG_PATH,
        help="The path to the Home Assistant configuration directory.",
    )
    parser.add_argument(
        "--package-file",
        "-p",
        type=Path,
        default=DEFAULT_PACKAGE_FILE,
        help="The path to the package file.",
    )
    parser.add_argument(
        "--git-tags",
        "-g",
        action="store_true",
        help="Use git to search for version tags. This will avoid GitHub API limits.",
    )

    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    # List installed packages
    list_parser = subparsers.add_parser("list", description="List installed packages.")
    list_parser.add_argument("--verbose", "-v", action="store_true")

    # List git tags for a given package
    list_tags_parser = subparsers.add_parser("tags", help="List tags for a package.")
    list_tags_parser.add_argument("url", type=str, help="The URL of the package.")
    list_tags_parser.add_argument(
        "--limit", type=int, default=10, help="The number of tags to display."
    )

    # Add packages
    add_parser = subparsers.add_parser("add", description="Add or install packages.")

    package_group = add_parser.add_mutually_exclusive_group(required=True)
    package_group.add_argument(
        "--file", "-f", type=Path, help="The path to a package file."
    )
    package_group.add_argument(
        "url", nargs="?", type=str, help="The URL of the package."
    )

    package_type_group = add_parser.add_mutually_exclusive_group()
    package_type_group.add_argument(
        "--integration",
        action="store_const",
        dest="type",
        const=PackageType.INTEGRATION,
        default=PackageType.INTEGRATION,
    )
    package_type_group.add_argument(
        "--plugin", action="store_const", dest="type", const=PackageType.PLUGIN
    )

    add_parser.add_argument(
        "--version", "-v", type=str, help="The version of the package."
    )
    add_parser.add_argument(
        "--update",
        "-u",
        action="store_true",
        help="Update the package if it already exists.",
    )
    add_parser.add_argument(
        "--ignore-versions",
        "-i",
        type=str,
        help="The version of the package to ignore. Multiple can be split by a comma.",
    )

    # Remove packages
    remove_parser = subparsers.add_parser(
        "remove", description="Remove installed packages."
    )
    remove_parser.add_argument("packages", nargs="+")

    # Upgrade packages
    update_parser = subparsers.add_parser(
        "upgrade", description="Upgrade installed packages."
    )
    update_parser.add_argument("packages", nargs="*")

    return parser


class Unhacs:
    def __init__(
        self,
        hass_config: Path = DEFAULT_HASS_CONFIG_PATH,
        package_file: Path = DEFAULT_PACKAGE_FILE,
    ):
        self.hass_config = hass_config
        self.package_file = package_file

    def read_lock_packages(self) -> list[Package]:
        return read_lock_packages(self.package_file)

    def write_lock_packages(self, packages: Iterable[Package]):
        return write_lock_packages(packages, self.package_file)

    def add_package(
        self,
        package_url: str,
        version: str | None = None,
        update: bool = False,
        package_type: PackageType = PackageType.INTEGRATION,
        ignore_versions: set[str] | None = None,
    ):
        """Install and add a package to the lock or install a specific version."""
        package = Package(
            package_url,
            version=version,
            package_type=package_type,
            ignored_versions=ignore_versions,
        )
        packages = self.read_lock_packages()

        # Raise an error if the package is already in the list
        existing_package = next((p for p in packages if p.url == package.url), None)
        if existing_package:
            if update:
                # Remove old version of the package
                packages = [p for p in packages if p.url != package.url]
            else:
                raise ValueError("Package already exists in the list")

        package.install(self.hass_config)

        packages.append(package)
        self.write_lock_packages(packages)

    def upgrade_packages(self, package_names: list[str]):
        """Uograde to latest version of packages and update lock."""
        installed_packages: Iterable[Package]

        if not package_names:
            installed_packages = get_installed_packages(self.hass_config)
        else:
            installed_packages = [
                p
                for p in get_installed_packages(self.hass_config)
                if p.name in package_names
            ]

        outdated_packages: list[Package] = []
        latest_packages = [p.get_latest() for p in installed_packages]
        for installed_package, latest_package in zip(
            installed_packages, latest_packages
        ):
            if latest_package != installed_package:
                print(
                    f"upgrade {installed_package.name} from {installed_package.version} to {latest_package.version}"
                )
                outdated_packages.append(latest_package)

        if outdated_packages and input("Upgrade all packages? (y/N) ").lower() != "y":
            return

        for installed_package in outdated_packages:
            installed_package.install(self.hass_config)

        # Update lock file to latest now that we know they are uograded
        latest_lookup = {p.url: p for p in latest_packages}
        packages = [latest_lookup.get(p.url, p) for p in self.read_lock_packages()]

        self.write_lock_packages(packages)

    def list_packages(self, verbose: bool = False):
        """List installed packages and their versions."""
        for package in get_installed_packages():
            print(package.verbose_str() if verbose else str(package))

    def list_tags(self, url: str, limit: int = 10):
        print(f"Tags for {url}:")
        for tag in get_repo_tags(url)[-1 * limit :]:
            print(tag)

    def remove_packages(self, package_names: list[str]):
        """Remove installed packages and uodate lock."""
        packages_to_remove = [
            package
            for package in get_installed_packages()
            if package.name in package_names
        ]
        remaining_packages = [
            package
            for package in self.read_lock_packages()
            if package not in packages_to_remove
        ]

        for package in packages_to_remove:
            package.uninstall(self.hass_config)

        self.write_lock_packages(remaining_packages)


def main():
    # If the sub command is add package, it should pass the parsed arguments to the add_package function and return
    parser = create_parser()
    args = parser.parse_args()

    unhacs = Unhacs(args.config, args.package_file)
    Package.git_tags = args.git_tags

    if args.subcommand == "add":
        # If a file was provided, update all packages based on the lock file
        if args.file:
            packages = read_lock_packages(args.file)
            for package in packages:
                unhacs.add_package(
                    package.url,
                    package.version,
                    update=True,
                    package_type=package.package_type,
                    ignore_versions=package.ignored_versions,
                )
        elif args.url:
            unhacs.add_package(
                args.url,
                version=args.version,
                update=args.update,
                package_type=args.type,
                ignore_versions=(
                    {version for version in args.ignore_versions.split(",")}
                    if args.ignore_versions
                    else None
                ),
            )
        else:
            raise ValueError("Either a file or a URL must be provided")
    elif args.subcommand == "list":
        unhacs.list_packages(args.verbose)
    elif args.subcommand == "tags":
        unhacs.list_tags(args.url, limit=args.limit)
    elif args.subcommand == "remove":
        unhacs.remove_packages(args.packages)
    elif args.subcommand == "upgrade":
        unhacs.upgrade_packages(args.packages)
    else:
        print(f"Command {args.subcommand} is not implemented")
        exit(1)


if __name__ == "__main__":
    main()



import json
import shutil
import tempfile
from collections.abc import Generator
from collections.abc import Iterable
from enum import StrEnum
from enum import auto
from io import BytesIO
from pathlib import Path
from typing import cast
from zipfile import ZipFile

import requests
import yaml

from unhacs.git import get_ref_zip
from unhacs.git import get_repo_tags

DEFAULT_HASS_CONFIG_PATH: Path = Path(".")
DEFAULT_PACKAGE_FILE = Path("unhacs.yaml")


def extract_zip(zip_file: ZipFile, dest_dir: Path):
    for info in zip_file.infolist():
        if info.is_dir():
            continue
        file = Path(info.filename)
        # Strip top directory from path
        file = Path(*file.parts[1:])
        path = dest_dir / file
        path.parent.mkdir(parents=True, exist_ok=True)
        with zip_file.open(info) as source, open(path, "wb") as dest:
            dest.write(source.read())


class PackageType(StrEnum):
    INTEGRATION = auto()
    PLUGIN = auto()


class Package:
    git_tags = False

    def __init__(
        self,
        url: str,
        version: str | None = None,
        package_type: PackageType = PackageType.INTEGRATION,
        ignored_versions: set[str] | None = None,
    ):
        self.url = url
        self.package_type = package_type
        self.ignored_versions = ignored_versions or set()

        parts = self.url.split("/")
        self.owner = parts[-2]
        self.name = parts[-1]

        self.path: Path | None = None

        if not version:
            self.version = self.fetch_version_release()
        else:
            self.version = version

    def __str__(self):
        return f"{self.name} {self.version}"

    def __eq__(self, other):
        return self.url == other.url and self.version == other.version

    def verbose_str(self):
        return f"{self.name} {self.version} ({self.url})"

    @staticmethod
    def from_yaml(yaml: dict) -> "Package":
        # Convert package_type to enum
        package_type = yaml.pop("package_type")
        if package_type and isinstance(package_type, str):
            package_type = PackageType(package_type)
            yaml["package_type"] = package_type

        return Package(**yaml)

    def to_yaml(self: "Package") -> dict:
        return {
            "url": self.url,
            "version": self.version,
            "package_type": str(self.package_type),
        }

    def add_ignored_version(self, version: str):
        self.ignored_versions.add(version)

    def _fetch_version_release_releases(self, version: str | None = None) -> str:
        # Fetch the releases from the GitHub API
        response = requests.get(
            f"https://api.github.com/repos/{self.owner}/{self.name}/releases"
        )
        response.raise_for_status()
        releases = response.json()

        if not releases:
            raise ValueError(f"No releases found for package {self.name}")

        # Default to latest
        desired_release = releases[0]

        # If a version is provided, check if it exists in the releases
        if version:
            for release in releases:
                if release["tag_name"] == version:
                    desired_release = release
                    break
            else:
                raise ValueError(f"Version {version} does not exist for this package")

        return cast(str, desired_release["tag_name"])

    def _fetch_version_release_git(self, version: str | None = None) -> str:
        tags = get_repo_tags(self.url)
        if not tags:
            raise ValueError(f"No tags found for package {self.name}")
        if version and version not in tags:
            raise ValueError(f"Version {version} does not exist for this package")

        tags = [tag for tag in tags if tag not in self.ignored_versions]
        if not version:
            version = tags[-1]

        return version

    def fetch_version_release(self, version: str | None = None) -> str:
        if self.git_tags:
            return self._fetch_version_release_git(version)
        else:
            return self._fetch_version_release_releases(version)

    def fetch_versions(self) -> list[str]:
        return get_repo_tags(self.url)

    def get_hacs_json(self, version: str | None = None) -> dict:
        """Fetches the hacs.json file for the package."""
        version = version or self.version
        response = requests.get(
            f"https://raw.githubusercontent.com/{self.owner}/{self.name}/{version}/hacs.json"
        )

        if response.status_code == 404:
            return {}

        response.raise_for_status()
        return response.json()

    def install_plugin(self, hass_config_path: Path):
        """Installs the plugin package."""

        valid_filenames: Iterable[str]
        if filename := self.get_hacs_json().get("filename"):
            valid_filenames = (cast(str, filename),)
        else:
            valid_filenames = (
                f"{self.name.removeprefix('lovelace-')}.js",
                f"{self.name}.js",
                f"{self.name}-umd.js",
                f"{self.name}-bundle.js",
            )

        def real_get(filename) -> requests.Response | None:
            urls = [
                f"https://raw.githubusercontent.com/{self.owner}/{self.version}/dist/{filename}",
                f"https://github.com/{self.owner}/{self.name}/releases/download/{self.version}/{filename}",
                f"https://raw.githubusercontent.com/{self.owner}/{self.version}/{filename}",
            ]

            for url in urls:
                plugin = requests.get(url)

                if int(plugin.status_code / 100) == 4:
                    continue

                plugin.raise_for_status()

                return plugin

            return None

        for filename in valid_filenames:
            plugin = real_get(filename)
            if plugin:
                break
        else:
            raise ValueError(f"No valid filename found for package {self.name}")

        js_path = hass_config_path / "www" / "js"
        js_path.mkdir(parents=True, exist_ok=True)
        js_path.joinpath(filename).write_text(plugin.text)

        yaml.dump(self.to_yaml(), js_path.joinpath(f"{filename}-unhacs.yaml").open("w"))

    def install_integration(self, hass_config_path: Path):
        """Installs the integration package."""
        zipball_url = get_ref_zip(self.url, self.version)
        response = requests.get(zipball_url)
        response.raise_for_status()

        with tempfile.TemporaryDirectory(prefix="unhacs-") as tempdir:
            tmpdir = Path(tempdir)
            extract_zip(ZipFile(BytesIO(response.content)), tmpdir)

            source, dest = None, None
            for custom_component in tmpdir.glob("custom_components/*"):
                source = custom_component
                dest = hass_config_path / "custom_components" / custom_component.name
                break
            else:
                hacs_json = json.loads((tmpdir / "hacs.json").read_text())
                if hacs_json.get("content_in_root"):
                    source = tmpdir
                    dest = hass_config_path / "custom_components" / self.name

            if not source or not dest:
                raise ValueError("No custom_components directory found")

            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.rmtree(dest, ignore_errors=True)
            shutil.move(source, dest)

            yaml.dump(self.to_yaml(), dest.joinpath("unhacs.yaml").open("w"))

    def install(self, hass_config_path: Path):
        """Installs the package."""
        if self.package_type == PackageType.PLUGIN:
            self.install_plugin(hass_config_path)
        elif self.package_type == PackageType.INTEGRATION:
            self.install_integration(hass_config_path)
        else:
            raise NotImplementedError(f"Unknown package type {self.package_type}")

    def uninstall(self, hass_config_path: Path) -> bool:
        """Uninstalls the package if it is installed, returning True if it was uninstalled."""
        if self.path:
            if self.path.is_dir():
                shutil.rmtree(self.path)
            else:
                self.path.unlink()
                self.path.with_name(f"{self.path.name}-unhacs.yaml").unlink()
            return True

        installed_package = self.installed_package(hass_config_path)
        if installed_package:
            installed_package.uninstall(hass_config_path)
            return True

        return False

    def installed_package(self, hass_config_path: Path) -> "Package|None":
        """Returns the installed package if it exists, otherwise None."""
        for package in get_installed_packages(hass_config_path, [self.package_type]):
            if package.url == self.url:
                return package

        return None

    def is_update(self, hass_config_path: Path) -> bool:
        """Returns True if the package is not installed or the installed version is different from the latest."""
        installed_package = self.installed_package(hass_config_path)
        return installed_package is None or installed_package.version != self.version

    def get_latest(self) -> "Package":
        """Returns a new Package representing the latest version of this package."""
        package = self.to_yaml()
        package.pop("version")
        return Package(**package)


def get_installed_packages(
    hass_config_path: Path = DEFAULT_HASS_CONFIG_PATH,
    package_types: Iterable[PackageType] = (
        PackageType.INTEGRATION,
        PackageType.PLUGIN,
    ),
) -> Generator[Package, None, None]:
    # Integration packages
    if PackageType.INTEGRATION in package_types:
        for custom_component in (hass_config_path / "custom_components").glob("*"):
            unhacs = custom_component / "unhacs.yaml"
            if unhacs.exists():
                package = Package.from_yaml(yaml.safe_load(unhacs.open()))
                package.path = custom_component
                yield package

    # Plugin packages
    if PackageType.PLUGIN in package_types:
        for js_unhacs in (hass_config_path / "www" / "js").glob("*-unhacs.yaml"):
            package = Package.from_yaml(yaml.safe_load(js_unhacs.open()))
            package.path = js_unhacs.with_name(
                js_unhacs.name.removesuffix("-unhacs.yaml")
            )
            yield package


# Read a list of Packages from a text file in the plain text format "URL version name"
def read_lock_packages(package_file: Path = DEFAULT_PACKAGE_FILE) -> list[Package]:
    if package_file.exists():
        return [
            Package.from_yaml(p)
            for p in yaml.safe_load(package_file.open())["packages"]
        ]
    return []


# Write a list of Packages to a text file in the format URL version name
def write_lock_packages(
    packages: Iterable[Package], package_file: Path = DEFAULT_PACKAGE_FILE
):
    yaml.dump({"packages": [p.to_yaml() for p in packages]}, package_file.open("w"))



import re
import subprocess
from dataclasses import dataclass


@dataclass
class GitTag:
    name: str
    version: tuple[int, int, int]
    suffix: str

    @staticmethod
    def parse(name: str):
        if result := re.match(r"^[v]?([\d.]+)(.*)", name):
            version_str = result.group(1)
            suffix = result.group(2)

            parts = version_str.split(".")
            if len(parts) > 3:
                raise ValueError(f"Invalid version tag: {name}")

            try:
                version = (
                    int(parts[0]),
                    int(parts[1]) if len(parts) > 1 else 0,
                    int(parts[2]) if len(parts) > 2 else 0,
                )
            except ValueError:
                raise ValueError(f"Invalid version tag: {name}")

            return GitTag(name, version, suffix)

    def __str__(self):
        return f"{self.name} {self.version}"

    def __eq__(self, other):
        return self.version == other.version and self.suffix == other.suffix

    def __lt__(self, other):
        return self.version < other.version or (
            self.version == other.version and self.suffix < other.suffix
        )


def get_repo_tags(repository_url: str) -> list[str]:
    # Run the command
    command = f"git -c 'versionsort.suffix=-' ls-remote --tags --sort='v:refname' {repository_url}"
    result = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    # Check for errors
    if result.returncode != 0:
        raise Exception(f"Error running command: {command}\n{result.stderr.decode()}")

    # Parse the output
    tags: list[GitTag] = []
    for line in result.stdout.decode().split("\n"):
        if line:
            if search_result := re.search(r"refs/tags/(.*)", line):
                tag = search_result.group(1)
                if git_tag := GitTag.parse(tag):
                    tags.append(git_tag)

    tags.sort()

    return [tag.name for tag in tags]


def get_ref_zip(repository_url: str, tag_name: str) -> str:
    return f"{repository_url}/archive/refs/tags/{tag_name}.zip"
